--[[
Restoration Shaman Stat Analyzer, by Rheeah

Expanded and enhanced version based on the original "Resto Shaman Mastery Analyzer" addon by _swift
See:
http://www.curse.com/addons/wow/rshamanmastery

--]]

local revision = tonumber(string.sub("$Revision: 34 $", 12, -3))
_G.RShamanStat = {
    ["revision"] = revision
}
local rss = _G.RShamanStat

-- item id's for gems that increase critical effect; only MoP gems that could be sensibly used by healers are included
local BURNING_PRIMAL_DIAMOND = 76885
local REVITALIZING_PRIMAL_DIAMOND = 76888
local criticalBonusGems = { BURNING_PRIMAL_DIAMOND, REVITALIZING_PRIMAL_DIAMOND }

-- item id's for T15 tier pieces 
local setBonusT15 = {     ["ChestSlot"] = { 95940 , 95310 , 96684 } , 
						  ["HandsSlot"] = { 95941 , 95311 , 96685 } , 
						   ["HeadSlot"] = { 95942 , 95312 , 96686 } ,
						   ["LegsSlot"] = { 95943 , 95313 , 96687 } , 
					   ["ShoulderSlot"] = { 95944 , 95314 , 96688 } }
-- item id's for Lei-Shen trinket
local lightningImbuedChalice = { ["Trinket0Slot"] = { 95817, 94530, 96189, 96561, 96933 } ,
                                 ["Trinket1Slot"] = { 95817, 94530, 96189, 96561, 96933 } }

-- spell id's for various heals; some are not needed (yet) and some item proc's that heal should probably be added
local HEALING_WAVE = 331
local HEALING_SURGE = 8004
local GREATER_HEALING_WAVE = 77472
local RIPTIDE = 61295
local ANCESTRAL_AWAKENING = 52752        -- the heal from AA; earlier versions had incorrect value!
local UNLEASH_LIFE = 73685
local CHAIN_HEAL = 1064
local HEALING_RAIN = 73921               -- the heal from Healing Rain
local EARTHLIVING = 51945
local EARTH_SHIELD = 379                 -- the heal from Earth Shield
local RESTORATIVE_MISTS = 114083         -- heal copy from Ascendance
local ANCESTRAL_GUIDANCE = 114911        -- heal copy from AG
local CLEANSING_WATERS = 55562			 -- heal from glyphed Purify Spirit
local GLYPH_OF_HEALING_WAVE = 55533      -- heal copy from Healing Wave
local FLAMES_OF_LIFE = 137808            -- heal from glyphed Flame Shock
local FLAME_SHOCK = 8050			     -- the damaging spell
local GIFT_OF_THE_NAARU = 59547          -- spacegoat racial
local SUMMON_HEALING_RAIN = 73920        -- the spell which casts Healing Rain
local SUMMON_HEALING_STREAM = 5394       -- the spell which summons HST
local SUMMON_HEALING_TIDE = 108280       -- the spell which summons HTT
local SUMMON_SPIRIT_LINK = 98008         -- the spell which summons SLT
local HEALING_STREAM = 52042             -- the actual heal applied by HST
local HEALING_STREAM_2T15 = 138304       -- the 'bonus' heal applied by HST with the T15 set bonus
local HEALING_TIDE = 114942              -- the actual heal applied by HTT
local SPIRIT_LINK = 98021          		 -- the actual heal applied by SLT
local PROC_RESTORATION_INFINITE = 139195 -- Lei Shen trinket proc
local HEALTHSTONE = 6262                 -- Warlock cookie
local WINDWOOL_BANDAGE = 102694
local HEAVY_WINDWOOL_BANDAGE = 102695
local MASTER_HEALING_POTION = 105708

local CONSUMABLES = { HEALTHSTONE , WINDWOOL_BANDAGE, HEAVY_WINDWOOL_BANDAGE, MASTER_HEALING_POTION } -- unaffected by Spellpower or Mastery
local BANDAGES = { WINDWOOL_BANDAGE , HEAVY_WINDWOOL_BANDAGE }                                        -- cannot crit


local RESURGENCE = 101033                -- Mana Gains are logged under combatEvent SPELL_ENERGIZE
-- some buffs we need to track
rss.AURA_TIDAL_WAVES = 53390
rss.AURA_ASCENDANCE = 114052
rss.AURA_ANCESTRAL_GUIDANCE = 108281
rss.AURA_EARTH_SHIELD = 974          
-- table to store unrecognized heals for debugging
local unknownSpells = { }

-- A table that links SpellIDs with their respective resurgence based mana gains
local RESURGENCE_GAINS = {
    [HEALING_WAVE] = 8849,
    [GREATER_HEALING_WAVE] = 8849,
    [HEALING_SURGE] = 8849 * 0.6,
    [RIPTIDE] = 8849 * 0.6,
    [UNLEASH_LIFE] = 8849 * 0.6,
    [CHAIN_HEAL] = 8849 * 0.333
}

-- A table that links SpellIDs with their healing power coefficients
local SPELL_COEFFICIENTS = {
	[HEALING_WAVE] = { ["SPELL_HEAL"] = { 7790, 8899, 0.756 } },
	[HEALING_SURGE] = { ["SPELL_HEAL"] = { 11687, 13351, 1.135 } },
	[GREATER_HEALING_WAVE] = { ["SPELL_HEAL"] = { 14172, 16190, 1.377 } },
	[RIPTIDE] = { ["SPELL_HEAL"] = { 3735, 3735, 0.339 } ,
	              ["SPELL_PERIODIC_HEAL"] = { 10584, 10584, 0.96 }},
	[UNLEASH_LIFE] = { ["SPELL_HEAL"] = { 3028, 3280, 0.286 } },
	[CHAIN_HEAL] = { ["SPELL_HEAL"] = { 7086, 8094, 0.6876 } },
	[HEALING_RAIN] = { ["SPELL_HEAL"] = { 2380, 2829, 0.2364 } },
	[EARTHLIVING] = { ["SPELL_PERIODIC_HEAL"] = { 4376, 4376, 0.3984 } },
	[EARTH_SHIELD] = { ["SPELL_HEAL"] = { 2043, 2043, 0.13 } },
	[HEALING_STREAM] = { ["SPELL_PERIODIC_HEAL"] = { 0, 0, 0.44 } },
	[HEALING_STREAM_2T15] = { ["SPELL_PERIODIC_HEAL"] = { 0, 0, 0.11 } },
	[HEALING_TIDE] = { ["SPELL_HEAL"] = { 4932, 4932, 0.484 } },
	[PROC_RESTORATION_INFINITE] = { ["SPELL_HEAL"] = { 0, 0, 1.25 } },
	[FLAMES_OF_LIFE] = { ["SPELL_DAMAGE"] = { 1086, 1086, 0.449 } ,
						 ["SPELL_PERIODIC_DAMAGE"] = { 2910, 2910, 2.10 } }
}

-- thankfully, this appears to be the only trinket with a healing proc that scales with spellpower
local chaliceBaseHeal =
{
	[502] = 46561,
	[506] = 48329,
	[510] = 50164,
	[522] = 56071,
	[526] = 58200,
	[528] = 56286,
	[530] = 60410,
	[532] = 58423,
	[535] = 63294,
	[536] = 60641,
	[539] = 65697,
	[541] = 63534,
	[543] = 68192,
	[545] = 65946,
	[549] = 68450
}

rss.intToCrit = 2533.66				-- Int to add +1% crit chance
rss.ratingToCrit = 600.0			-- Crit Rating to add +1% crit
rss.ratingToMastery = 200.0         -- Mastery Rating to add +1% mastery

function rss.ResurgenceGains(record)
    local sum = 0
    for spellid, gain in pairs(RESURGENCE_GAINS) do
        if record.spells[spellid] then
            sum = sum + record.spells[spellid] * gain
        end
    end
    return sum
end


local aaTriggers = { HEALING_SURGE, HEALING_WAVE, GREATER_HEALING_WAVE, RIPTIDE, UNLEASH_LIFE }

local criticalMultiplier = 2.0           -- multiplier for critical effects; default value is 2
local aaSetBonusT15 = false              -- flag to indicate whether the 4T15 set bonus is applicable

local lastCast = { }                     -- table of most recent cast of each spell (indexed by spell ID), for matching with AA
local healthTable = { }                  -- table of most recent health and max health of every known friendly unit
local spiritLinkGUID = nil               -- ID of our currently summoned Spirit Link totem, or nil

local fight = nil;
local playerState = nil;


-- check for a crit-enhancing metagem
function RShamanStatCheckMeta()
	local gem1, gem2, gem3 = GetInventoryItemGems(GetInventorySlotInfo("HeadSlot")); -- head slot
	if tContains(criticalBonusGems, gem1) or tContains(criticalBonusGems, gem2) or tContains(criticalBonusGems, gem3) then
		criticalMultiplier = 2.0 * 1.03;
	else
		criticalMultiplier = 2.0;
	end
end

-- Find the item level of something equipped in a slot (including upgrades, which is why this isn't as easy as it should be)
function RShamanStatGetItemLevel(slot)
	local g=MyTooltipScan or CreateFrame("GameTooltip","MyTooltipScan",UIParent,"GameTooltipTemplate")
	g:SetOwner(UIParent,"ANCHOR_NONE")
	g:SetInventoryItem("player",slot)
	for i=1,g:NumLines() do
		local line = _G["MyTooltipScanTextLeft"..i]:GetText();
		local ilvl = line:match('%d%d%d$');
		if ilvl then
			return tonumber(ilvl);
		end
	end
end

-- Scan equipped gear for tier 15 pieces and the Lei-Shen trinket
function RShamanStatCheckGear()
	aaSetBonusT15 = false;
	local n15 = 0;
	for slotName, tierItems in pairs(setBonusT15) do
		local equipID = GetInventoryItemID("player", GetInventorySlotInfo(slotName));
		if equipID and tContains(tierItems, equipID) then
			n15 = n15 + 1;
		end
	end
	if n15 >= 4 then
		aaSetBonusT15 = true;
	end
	
	local chaliceLevel = nil;
	for slotName, trinkets in pairs(lightningImbuedChalice) do
		local equipID = GetInventoryItemID("player", GetInventorySlotInfo(slotName));
		if equipID and tContains(trinkets, equipID) then
			chaliceLevel = RShamanStatGetItemLevel(GetInventorySlotInfo(slotName));
			break;
		end
	end
	-- fill the appropriate base heal determined by the item level
	if chaliceLevel and chaliceBaseHeal[chaliceLevel] then
		local baseHeal = chaliceBaseHeal[chaliceLevel];
		local coeff = SPELL_COEFFICIENTS[PROC_RESTORATION_INFINITE]["SPELL_HEAL"];
		coeff[1] = baseHeal;
		coeff[2] = baseHeal;
	end
end

-- initialization
function RShamanStatOnLoad(self)
	local lClass, eClass = UnitClass("player");
	if eClass ~= "SHAMAN" then
		return;
	end
	-- enable escape to close the window
	tinsert(UISpecialFrames, "RShamanStatFrame")
	
	-- listen for combat log and other events
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self:RegisterEvent("PLAYER_REGEN_ENABLED");
	self:RegisterEvent("PLAYER_REGEN_DISABLED");
	self:RegisterEvent("UNIT_MAXHEALTH");
	self:RegisterEvent("UNIT_HEALTH_FREQUENT");
	self:RegisterEvent("PLAYER_TARGET_CHANGED");
	self:RegisterEvent("PLAYER_FOCUS_CHANGED");
	self:RegisterEvent("MASTERY_UPDATE");
	self:RegisterEvent("PLAYER_DAMAGE_DONE_MODS");
	self:RegisterEvent("COMBAT_RATING_UPDATE");
	self:RegisterEvent("UNIT_AURA");
	self:RegisterEvent("UNIT_STATS");
	self:RegisterEvent("VARIABLES_LOADED");
	RShamanStatClear();
	print(("Resto Shaman Stat Analyzer (r%s) loaded!"):format(rss.revision));
end


function RShamanStatPrintRSM()
    if (#RShamanStatArchive == 0) then
        print("RSS: no data available");
        return;
    end

    for i, v in ipairs(RShamanStatArchive) do
        if (RShamanStatArchive.filter.useAll or v.isBoss) then
            if v.effectiveTotal ~= 0 and v.effectiveSubTotal ~= 0 then
                print(("RSS: #%02d - %s: Healing Done: %.1fk, Mastery Effect: %.2f"):format(i, v.encounterName, v.effectiveTotal/1000, v.weightedTotal/v.effectiveSubTotal));
            else
                print(("RSS: #%02d - %s: No Healing Done"):format(i, v.encounterName));
            end
        end
    end
end


-- erase current combat data AND all archived/saved data
function RShamanStatClearAll()
	unknownSpells = { };
    RShamanStatClear();
    RShamanStatArchive = rss.CreateArchive();
end

-- erase current combat data; called when entering or leaving combat
function RShamanStatClear()
    lastCast = { };
	if UnitHealthMax("player") > 0 then
		healthTable = { [UnitGUID("player")] = { maxHealth = UnitHealthMax("player") , currentHealth = UnitHealth("player") } };
	else
		healthTable = { };
	end
    spiritLinkGUID = nil;
	if playerState then
		playerState.ascendance = nil;
		playerState.ancestralGuidance = nil;
	end
    fight = rss.CreateRecord();
	playerState = rss.MonitorState(fight);

    RShamanStatCheckMeta();
    RShamanStatCheckGear();
end


-- Print data from combat that just ended
-- Save it to archive table
-- Zero accumulators for next combat
function RShamanStatReset()
	if fight.effectiveTotal > 0 then
		if playerState and playerState.ascendance then
			rss.ProcessCopyRecord(playerState.ascendance, fight);
		end
		if playerState and playerState.ancestralGuidance then
			rss.ProcessCopyRecord(playerState.ancestralGuidance, fight);
		end
        rss.SaveRecord(RShamanStatArchive, fight)
        rss.PrintRecord(fight)
    end
	RShamanStatClear()
end


SLASH_RSHAMANSTAT1 = '/rss';

function RShamanStatPrintDifficulty()
    if (RShamanStatArchive.filter.difficulty ~= 0) then
        local difficultyName = GetDifficultyInfo(RShamanStatArchive.filter.difficulty);
        print("Instance difficulty is set to '"..difficultyName.."'");
    else
        print("Instance difficulty is set to 'Any'");
    end
end

function RShamanStatPrintUseAll()
    if (RShamanStatArchive.filter.useAll) then
        print("Recording all encounters.");
    else
        print("Recording boss encounters only.");
    end
end

-- Supported commands:
function RShamanStatCmd(msg, ...)
    local msg, args = msg:match("^(%S*)%s*(.-)$");
    if (msg == "help") then
        print(("Resto Shaman Stats (r%s) - Available Commands:"):format(revision));
        print(" * /rss sum          Print all data, and grand totals");
        print(" * /rss sum <boss>   Print data for the specified boss");
        print(" * /rss clr            Print all data, and grand totals, then wipe all data");
        print(" * /rss boss          Only include boss encounter data for grand totals");
        print(" * /rss all            Include all data for grand totals");
        print(" * /rss rsm            Print data of last encounter in RSM mode");
        print(" * /rss difficulty <Any,10N,10H,25H,25N,LFR> Print data from a specific difficulty only");


        RShamanStatPrintUseAll()
        RShamanStatPrintDifficulty()
        return;
    end
	
	if (msg == "") then
		RShamanStatFrame:Show()
	end

    if (msg == "rsm") then
        RShamanStatPrintRSM();
        return;
    end
	
	if msg == "debug" then
		local found;
		for id, spell in pairs(unknownSpells) do
			found = true;
			print("Unrecognized: "..spell.name.."("..id..") : "..spell.count);
		end
		if not found then
			print("No unrecognized spells encountered!");
		end
		
        local filter = function(r)
            return (RShamanStatArchive.filter.useAll or r.isBoss) and (RShamanStatArchive.filter.difficulty == 0 or RShamanStatArchive.filter.difficulty == r.difficulty);
        end
        local summary = rss.CreateAggregateRecord("Summary", RShamanStatArchive, filter);
		if (summary.effectiveTotal > 0) then
			for reason, leak in pairs(summary.leakage) do
				print(("%s : %.1f  (%.2f%%)"):format(reason, leak, 100*leak/summary.effectiveTotal));
			end
		end
	end

    if (msg == "sum" and args ~= "") then
        local r = rss.CreateAggregateRecord(args, RShamanStatArchive, function(r)
            return (r.encounterName == args) and (RShamanStatArchive.filter.difficulty == 0 or RShamanStatArchive.filter.difficulty == r.difficulty);
        end)
        rss.args = args
        rss.PrintRecord(r)
        return;
    end

    if (msg == "difficulty" and args ~= "") then
        if args == "10N" then
            RShamanStatArchive.filter.difficulty = 3;
        elseif args == "25N" then
            RShamanStatArchive.filter.difficulty = 4;
        elseif args == "10H" then
            RShamanStatArchive.filter.difficulty = 5;
        elseif args == "25H" then
            RShamanStatArchive.filter.difficulty = 6;
        elseif args == "LFR" then
            RShamanStatArchive.filter.difficulty = 7;
        else
            RShamanStatArchive.filter.difficulty = 0;
        end
        RShamanStatPrintDifficulty()
        return;
    end

    if (msg == "boss") then
        RShamanStatArchive.filter.useAll = false;
        RShamanStatPrintUseAll()
        return
	elseif (msg == "all") then
        RShamanStatArchive.filter.useAll = true;
        RShamanStatPrintUseAll()
        return
    end

	if (msg == "sum" or msg == "clr") then
        local filter = function(r)
            return (RShamanStatArchive.filter.useAll or r.isBoss) and (RShamanStatArchive.filter.difficulty == 0 or RShamanStatArchive.filter.difficulty == r.difficulty)
        end
		if (#RShamanStatArchive > 0) then
            for i, v in ipairs(RShamanStatArchive) do
                if (filter(v)) then
                    rss.PrintRecord(v)
                end
			end
        end

        local summary = rss.CreateAggregateRecord("Summary", RShamanStatArchive, filter)
        rss.PrintRecord(summary)
		print("---------------");
	end
	if (msg == "clr") then
		RShamanStatClearAll();
		print("RSS:: All archived data cleared");
	end
end
SlashCmdList["RSHAMANSTAT"] = RShamanStatCmd;

-- returns the name and classification of any hostile target of anyone we heal (boss detection)
function RShamanStatCheckTarget(unit)
	local unitTarget = unit.."-target";
	if ( UnitExists(unitTarget) and UnitAffectingCombat(unitTarget) and UnitCanAttack(unit, unitTarget) ) then
		return UnitName(unitTarget), UnitClassification(unitTarget);
	end
end

-- sets boss flag if worldboss target is detected, and sets encounter name
function RShamanStatCheckBoss(unit)
	local name, classification = RShamanStatCheckTarget(unit);
	if (classification == "worldboss" or (name ~= nil and fight.encounterName == "Unknown")) then
		fight.encounterName = name;
	end
	if (classification == "worldboss") then
		fight.isBoss = true;
		print("RSS:: Detected boss encounter "..name);
	end
end

-- matches an AA with the spell that triggered it, based on closest match to 30%
function RShamanStatMatchAA(amount)
	local mindiff = amount;                                         -- Match this AA to the spell that triggered it, based on the 30% relationship
	local sMin = nil;
	local vMin = nil;
	for s, v in pairs(lastCast) do
		if math.abs(v.amount*0.3 - amount) < mindiff then
			mindiff = math.abs(v.amount*0.3 - amount);
			vMin = v;
			sMin = s;
		end
	end
	return sMin, vMin;
end

function RShamanStatCalcCopy(record, amount, incrementalIntGain, incrementalMasteryGain, critical, critBonusAmount)
	record.rawHealing = record.rawHealing + amount;
	record.incrementalInt = record.incrementalInt + incrementalIntGain;
	record.incrementalMastery = record.incrementalMastery + incrementalMasteryGain;
	if critical then
		record.incrementalCritical[1] = record.incrementalCritical[1] + (0.01/rss.ratingToCrit)*critBonusAmount;
	else
		record.incrementalCritical[2] = record.incrementalCritical[2] + (0.01/rss.ratingToCrit)*critBonusAmount;
	end
end

function RShamanStatLeak(reason, healed)
	print("Leakage!");
	if not fight.leakage[reason] then
		fight.leakage[reason] = 0.0;
	end
	fight.leakage[reason] = fight.leakage[reason] + healed;
end

local lastFlameShock = {combatEvent = "SPELL_DAMAGE" , critical = nil};
local spellSchools = { }; -- lookup table to cache expensive logarithm calculation 

function RShamanStatCombatLogEntry(...)
	local playerGUID = UnitGUID("player");
	local timestamp, combatEvent, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags = select(1, ...);
	
	if sourceGUID == playerGUID and combatEvent and strfind(combatEvent, 'SPELL_SUMMON') then
		local spellId, spellName, spellSchool = select(12, ...);
		if spellId == SUMMON_SPIRIT_LINK then 
			spiritLinkGUID = destGUID;
		end
		return;
	end
	
	if combatEvent and strfind(combatEvent, 'UNIT_DESTROYED') and destGUID == spiritLinkGUID then
		spiritLinkGUID = nil;
		return;
    end

    -- handle power gains like resurgence
    if sourceGUID == playerGUID and combatEvent and combatEvent == "SPELL_ENERGIZE" then
        local spellId, spellName, spellSchool = select(12, ...)
        local amount, powerType = select(15, ...)
        if spellId == RESURGENCE then
            fight.resurgence = fight.resurgence + amount
        end
        return;
    end
	
	-- deal with the Flame Shock glyph
	if sourceGUID == playerGUID and combatEvent and strfind(combatEvent, '_DAMAGE') then
		local spellId = select(12,...);
		if spellId == FLAME_SHOCK then
			lastFlameShock.combatEvent = combatEvent;
			lastFlameShock.critical = select(21,...);
		end
	end

	local spiritLinkHeal = spiritLinkGUID and sourceGUID == spiritLinkGUID;
	if (sourceGUID == playerGUID or spiritLinkHeal) and combatEvent and strfind(combatEvent, '_HEAL') and healthTable[destGUID] then		
		-- Check for boss encounter
		if ( (not fight.isBoss) or (fight.encounterName == "Unknown") ) then
			RShamanStatCheckBoss(destName);
		end
		-- Check for first action of combat
		if fight.startTime <= 0 then
            fight.startTime = timestamp;
		end
        fight.totalTime = timestamp - fight.startTime;
		
		-- We healed somebody
		local spellId, spellName, spellSchool = select(12, ...)
		local amount, overhealing, absorbed, critical = select(15, ...)
		local healed = math.max(0.,amount - overhealing);                       -- ignore overhealing
        fight.effectiveTotal = fight.effectiveTotal + healed;                   -- accumulate total effective throughput
		
		if spiritLinkHeal then 													-- stat effects on Spirit Link totem heals doubtful
			return;
        end

        -- count the number of times we healed with non-periodic spells
        if combatEvent == "SPELL_HEAL" then
            local spellId_count = fight.spells[spellId] or 0
            spellId_count = spellId_count + 1
            fight.spells[spellId] = spellId_count
        end

		-- match AA once for everything that follows
		local aaSpell = nil;
		local aaCast  = nil;
		if spellId == ANCESTRAL_AWAKENING then
			aaSpell, aaCast = RShamanStatMatchAA(amount);
			if aaCast == nil then
				RShamanStatLeak("AA match not found", healed);
				return;
			end
		end
		
		-- deal with spell copies from Ascendance and Ancestral Guidance
		if spellId == RESTORATIVE_MISTS or spellId == ANCESTRAL_GUIDANCE then
			if playerState.ascendance and spellId == RESTORATIVE_MISTS then
				playerState.ascendance.effectiveCopies = playerState.ascendance.effectiveCopies + healed;
			elseif playerState.ancestralGuidance and spellId == ANCESTRAL_GUIDANCE then
				playerState.ancestralGuidance.effectiveCopies = playerState.ancestralGuidance.effectiveCopies + healed;	
			else
				RShamanStatLeak("Missing originals for copy", healed);  -- derp, we have no record of the spell(s) that this heal was copied from
			end
			return; -- nothing more to do for now; stats affect the spell that was copied, not the copy itself
		end
		
		-- Recognize consumables so they can be (mostly) ignored
		local consumable = tContains(CONSUMABLES, spellId);
		local bandage;
		if consumable then
			bandage = tContains(BANDAGES, spellId);
		end
		
		-- Deal with Intellect
		local mySpellPower = playerState.healingBonus;
		local incrementalInt = 0.0;		
		if SPELL_COEFFICIENTS[spellId] and SPELL_COEFFICIENTS[spellId][combatEvent] then
			local c = SPELL_COEFFICIENTS[spellId][combatEvent];
			local baseAmt = (c[1]+c[2])/2 + mySpellPower*c[3];
			incrementalInt = (1.1*c[3]/baseAmt) * amount;   -- the 1.1 reflects our 10% spellpower buff
		elseif spellId == ANCESTRAL_AWAKENING then
			incrementalInt = 0.3*aaCast.incrementalIntellect;
		elseif spellId == GLYPH_OF_HEALING_WAVE then
			if lastCast[HEALING_WAVE] then
				incrementalInt = 0.2*lastCast[HEALING_WAVE].incrementalIntellect;
			else
				RShamanStatLeak("Missing original HW for glyph heal", healed);
				return;
			end
		elseif spellId == FLAMES_OF_LIFE then
			mySpellPower = playerState.damageBonus[3];
			local c = SPELL_COEFFICIENTS[spellId][lastFlameShock.combatEvent];
			local baseAmt = (c[1]+c[2])/2 + mySpellPower*c[3];
			incrementalInt = (1.1*c[3]/baseAmt) * amount;
		elseif spellId ~= GIFT_OF_THE_NAARU and spellId ~= CLEANSING_WATERS and not consumable then
			if not unknownSpells[spellId] then
				print("Unrecognized healing entry: "..spellName.." "..spellId.." "..combatEvent);
				unknownSpells[spellId] = { name = spellName , count = 0 };
			end
			unknownSpells[spellId].count = unknownSpells[spellId].count + 1;
			consumable = true;  -- assume no stat benefits other than crit
		end
		if overhealing == 0 then
			fight.incrementalIntellect = fight.incrementalIntellect + incrementalInt;
		end
		
		-- Deal with Mastery
		local masteryGain = 0.0;
		local incrementalMasteryGain = 0.0;
		local effectiveMasteryGain = 0.0;
		local hpratio = 0.0;
		local unitHealthTable = healthTable[destGUID];							-- find last known values of health and max health
		if spellId ~= FLAMES_OF_LIFE and spellId ~= GIFT_OF_THE_NAARU and spellId ~= PROC_RESTORATION_INFINITE and not consumable then
			if spellId == ANCESTRAL_AWAKENING then
				masteryGain = 0.3 * aaCast.masteryGain;
				effectiveMasteryGain = math.max(0.0, masteryGain - overhealing);
				incrementalMasteryGain = 0.3 * aaCast.incrementalMastery;
				hpratio = aaCast.hpRatio;
			elseif spellId == GLYPH_OF_HEALING_WAVE then
				masteryGain = 0.2 * lastCast[HEALING_WAVE].masteryGain;
				effectiveMasteryGain = math.max(0.0, masteryGain - overhealing);
				incrementalMasteryGain = 0.2 * lastCast[HEALING_WAVE].incrementalMastery;
				hpratio = lastCast[HEALING_WAVE].hpRatio;			
			else
				local myMasteryBonus = playerState.masteryBonus;
				hpratio = 1 - (unitHealthTable.currentHealth/unitHealthTable.maxHealth);
				local healBaseAmount = amount/(1+hpratio*myMasteryBonus);               -- the size of the heal with 0% Mastery
				masteryGain = amount - healBaseAmount;                            		-- increase from base amount from Mastery
				effectiveMasteryGain = math.max(0.0, masteryGain - overhealing);  		-- overhealing is not useful
				incrementalMasteryGain = (0.01/rss.ratingToMastery) * hpratio * healBaseAmount;     -- how much would +1 Mastery Rating increase this heal?
			end
			fight.weightedTotal = fight.weightedTotal + healed * hpratio
			fight.effectiveSubTotal = fight.effectiveSubTotal + healed
			fight.masteryTotal = fight.masteryTotal + effectiveMasteryGain;						-- accumulate gain from current Mastery
			if (overhealing == 0) then
				fight.incrementalMastery = fight.incrementalMastery + incrementalMasteryGain;       -- accumulate gain from +1 Mastery
			end	
		end
		
		-- Deal with Crit
		local schoolIndex;
		if spellId == FLAMES_OF_LIFE then
			schoolIndex = 3;
			critical = lastFlameShock.critical;
		else
			if not spellSchools[spellSchool] then
				spellSchools[spellSchool] =  math.floor(1+log(spellSchool)/log(2));
			end
			schoolIndex = spellSchools[spellSchool];
		end
		
		local myCriticalChance = playerState.criticalChance[schoolIndex] or playerState.criticalChance[4];
		local thisCast = nil;                                                   -- record this as the most recent cast of this spell, for matching with AA
		if spellId ~= ANCESTRAL_AWAKENING then                                  -- unless it is AA healing, which can't trigger AA
			thisCast = { };
			thisCast.amount = amount;
			thisCast.hpRatio = hpratio;
			thisCast.critChance = myCriticalChance;
			thisCast.spellPower = mySpellPower;
			thisCast.masteryGain = masteryGain;
			thisCast.incrementalMastery = incrementalMasteryGain;
			thisCast.incrementalIntellect = incrementalInt;
			thisCast.critical = false;
		end
		local critBonusAmount = 0.0;
		if (critical or (spellId == ANCESTRAL_AWAKENING) or (spellId == GLYPH_OF_HEALING_WAVE and lastCast[HEALING_WAVE].critical) ) then                  
			if spellId == HEALING_SURGE then                                    -- is this a Healing Surge critical?
				if playerState.tidalWaves then              					-- tested: even if this cast consumed the last TW stack, we still have the buff at this moment
					myCriticalChance = math.min(1.0, myCriticalChance + 0.3);   -- adjust HS critical chance for TW
					thisCast.critChance = myCriticalChance;                     -- record the buffed value of critical chance for use with any AA triggered by this heal
				end
			end
			if (critical or (spellId == GLYPH_OF_HEALING_WAVE and lastCast[HEALING_WAVE].critical) ) then   -- this cast crit, so figure out how much we gained over the base heal
				thisCast.critical = true;                                       -- record the cast as critical so we can add the triggered AA to the critical category
				local critBaseAmount = amount/criticalMultiplier;               
				critBonusAmount = amount - critBaseAmount;
                fight.criticalTotal = fight.criticalTotal + math.max(0.0, critBonusAmount - overhealing);  -- record the effective gain from this crit
				-- how much would +1 Critical Rating increase the *expected* (average) effective amount of this cast?
				if (myCriticalChance > 0) then
                    fight.incrementalCritical[1] = fight.incrementalCritical[1] + (0.01/rss.ratingToCrit)*math.max(0.0, critBonusAmount - overhealing)/myCriticalChance;
					critBonusAmount = critBonusAmount/myCriticalChance;
				end
			else -- Ancestral Awakening                                      
				if aaCast ~= nil then
					if aaCast.critical then                          			   -- if the trigger for this AA was a crit (rather than the set bonus)..
					    local nonCritAA = 0;                                       -- size of a possible non-crit AA from original heal; zero unless set bonus present
						if aaSetBonusT15 then
							nonCritAA = 0.5 * 0.3 * aaCast.amount / criticalMultiplier     -- with set bonus, half of 30% of the non-crit heal amount expected
						end
                        fight.criticalTotal = fight.criticalTotal + math.max(0, healed - nonCritAA); -- count it as gain from critical
						if (aaCast.critChance > 0) then
							-- how much would +1 Critical Rating increase the expected (average) value of the cast that triggered it?
                            fight.incrementalCritical[1] = fight.incrementalCritical[1] + (0.01/rss.ratingToCrit)*math.max(0, healed - nonCritAA)/aaCast.critChance;
						end
					else  -- this AA was triggered by the set bonus; remove the current (non-crit) AA amount from the gain we added previously
                        fight.incrementalCritical[2] = math.max(0, fight.incrementalCritical[2] - (0.01/rss.ratingToCrit)*healed);
					end
				end
			end
		elseif not bandage then -- this is not a critical or an AA
			local criticalAmount = amount * criticalMultiplier;                -- how big would the raw heal be if it crit?
			critBonusAmount = criticalAmount - amount;
			if overhealing == 0 then										   -- no benefit to the target from critting on an overheal
				local criticalHeal = math.min(criticalAmount, unitHealthTable.maxHealth - unitHealthTable.currentHealth);
                fight.incrementalCritical[2] = fight.incrementalCritical[2] + (0.01/rss.ratingToCrit)*(criticalHeal - healed);
			end
			-- add healing from the hypothetical AA, if applicable; we assume it would hit the friendly on our list with the largest health deficit ("most injured")
			if tContains(aaTriggers, spellId) and (spellId ~= RIPTIDE or strfind(combatEvent, 'PERIODIC_HEAL') == nil) then
				local criticalAA = 0.3 * criticalAmount;                           -- how big would the AA heal be if it crit?
				local maxDeficit = 0;
				for guid, t in pairs(healthTable) do
					local deficit = math.max(0, t.maxHealth - t.currentHealth);
					if guid == destGUID then                                  -- account for the amount we *would* have healed the target...
						deficit = math.max(0, deficit - criticalAmount);
					end
					maxDeficit = math.max(maxDeficit, deficit);
				end
				criticalAA = math.min(criticalAA, maxDeficit);
                fight.incrementalCritical[2] = fight.incrementalCritical[2] + (0.01/rss.ratingToCrit)*criticalAA;
			end
		end
		if (thisCast ~= nil and tContains(aaTriggers, spellId) ) then         -- store information on this cast for AA matching and accounting
			lastCast[spellId] = thisCast;
		end
	-- done with Mastery and Crit, now store info for active copy buffs
		if (playerState.ascendanceAura or playerState.guidanceAura) and spellId ~= ANCESTRAL_AWAKENING then
			if playerState.ascendance then
				RShamanStatCalcCopy(playerState.ascendance, amount, incrementalInt, incrementalMasteryGain, critical, critBonusAmount);
			end
			if playerState.ancestralGuidance and combatEvent == "SPELL_HEAL" and spellId ~= GLYPH_OF_HEALING_WAVE then
				RShamanStatCalcCopy(playerState.ancestralGuidance, amount, incrementalInt, incrementalMasteryGain, critical, critBonusAmount)
			end
		end
	end
end
	
function RShamanStatOnEvent(self, event, ...)
	local currentSpec = GetSpecialization()
	local role = currentSpec and select(6, GetSpecializationInfo(currentSpec)) or "NONE"

    if event == "VARIABLES_LOADED" then
        RShamanStatArchive = rss.LoadArchive(RShamanStatArchive)
        if RShamanStatArchive and #RShamanStatArchive > 0 then
            DEFAULT_CHAT_FRAME:AddMessage("RSS: "..#RShamanStatArchive.." archived records found.");
        end
		rss.UpdateUI();
		-- fill in health table for ourselves, at least
		local playerGUID = UnitGUID("player");
		if playerGUID then
			local unitHealthTable = healthTable[playerGUID];
			if unitHealthTable == nil then
				unitHealthTable = { };
				healthTable[playerGUID] = unitHealthTable;
			end
			unitHealthTable.maxHealth = UnitHealthMax("player");
			unitHealthTable.currentHealth = UnitHealth("player");
			if unitHealthTable.maxHealth < 1 then
				healthTable[playerGUID] = nil;
			end
		end
        return;
    end
	if role ~= "HEALER" then
		return;
    end

	if event == "UNIT_HEALTH_FREQUENT" or event == "UNIT_MAXHEALTH" then
		local unitID = select(1,...);
		if unitID == nil then return; end
		if UnitIsPlayer(unitID) or UnitPlayerOrPetInParty(unitID) or UnitPlayerOrPetInRaid(unitID) then
			local unitGUID = UnitGUID(unitID);
			if unitGUID == nil then return; end
			local unitHealthTable = healthTable[unitGUID];
			if unitHealthTable == nil then
				unitHealthTable = { };
				healthTable[unitGUID] = unitHealthTable;
			end
			unitHealthTable.maxHealth = UnitHealthMax(unitID);
			unitHealthTable.currentHealth = UnitHealth(unitID);
			if unitHealthTable.maxHealth < 1 then
				healthTable[unitGUID] = nil;
			end
		end
		return;
	end
	
	if event == "PLAYER_TARGET_CHANGED" then
		if UnitCanAssist("player","target") then
			local unitGUID = UnitGUID("target");
			if not unitGUID then return; end
			local unitHealthTable = healthTable[unitGUID];
			if unitHealthTable == nil then
				unitHealthTable = { };
				healthTable[unitGUID] = unitHealthTable;
			end
			unitHealthTable.maxHealth = UnitHealthMax("target");
			unitHealthTable.currentHealth = UnitHealth("target");
			if unitHealthTable.maxHealth < 1 then
				healthTable[unitGUID] = nil;
			end
		end
		return;
	end
	
	if event == "PLAYER_FOCUS_CHANGED" then
		if UnitCanAssist("player","focus") then
			local unitGUID = UnitGUID("focus");
			if not unitGUID then return; end
			local unitHealthTable = healthTable[unitGUID];
			if unitHealthTable == nil then
				unitHealthTable = { };
				healthTable[unitGUID] = unitHealthTable;
			end
			unitHealthTable.maxHealth = UnitHealthMax("focus");
			unitHealthTable.currentHealth = UnitHealth("focus");
			if unitHealthTable.maxHealth < 1 then
				healthTable[unitGUID] = nil;
			end
		end
		return;
	end
	
	if event == "MASTERY_UPDATE" or event == "PLAYER_DAMAGE_DONE_MODS" or event == "COMBAT_RATING_UPDATE" or 
	   (event == "UNIT_AURA" and UnitIsUnit(select(1,...),"player")) or (event == "UNIT_STATS" and UnitIsUnit(select(1,...),"player")) then
		playerState = rss.MonitorState(fight);
	end
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		RShamanStatCombatLogEntry(...);
		return;
	end
	if event == "PLAYER_REGEN_DISABLED" then
		-- entering combat; reset counters
		RShamanStatClear();
		return;
	end
	if event == "PLAYER_REGEN_ENABLED" then
		-- leaving combat: print status and reset counters
		RShamanStatReset();
		return;
	end
end